# -*- coding: utf-8 -*-


from .parser import parse
from .utils import transpose_note, display_appended, display_on, note_to_val


class Chord(object):
    """ Class to handle a chord.

    :param str _chord: Name of the chord. (e.g. C, Am7, F#m7-5/A)
    :param str _root: The root note of chord.
    :param pychord.Quality _quality: The quality of chord. (e.g. m7, 6, M9, ...)
    :param list[str] _appended: The appended notes on chord.
    :param str _on: The base note of slash chord.
    """
    def __init__(self, chord):
        """ Constructor of Chord instance

        :param str chord: Name of chord.
        """
        self._chord = chord
        self._root, self._quality, self._appended, self._on = "", "", "", ""
        self._parse(chord)

    def __unicode__(self):
        return self._chord

    def __str__(self):
        return self._chord

    def __repr__(self):
        return "<Chord: {}>".format(self._chord)

    def __eq__(self, other):
        if not isinstance(other, Chord):
            raise TypeError("Cannot compare Chord object with {} object".format(type(other)))
        if note_to_val(self._root) != note_to_val(other.root):
            return False
        if self._quality != other.quality:
            return False
        if self._appended != other.appended:
            return False
        if self._on and other.on:
            if note_to_val(self._on) != note_to_val(other.on):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def chord(self):
        """ The name of chord """
        return self._chord

    @property
    def root(self):
        """ The root note of chord """
        return self._root

    @property
    def quality(self):
        """ The quality of chord """
        return self._quality

    @property
    def appended(self):
        """ The appended notes on chord """
        return self._appended

    @property
    def on(self):
        """ The base note of slash chord """
        return self._on

    def info(self):
        """ Return information of chord to display """
        return """{}
root={}
quality={}
appended={}
on={}""".format(self._chord, self._root, self._quality, self._appended, self._on)

    def transpose(self, trans, scale="C"):
        """ Transpose the chord

        :param int trans: Transpose key
        :param str scale: key scale
        :return:
        """
        if not isinstance(trans, int):
            raise TypeError("Expected integers, not {}".format(type(trans)))
        self._root = transpose_note(self._root, trans, scale)
        if self._on:
            self._on = transpose_note(self._on, trans, scale)
        self._reconfigure_chord()

    def components(self, visible=True):
        """ Return the component notes of chord

        :param bool visible: returns the name of notes if True else list of int
        :rtype: list[(str or int)]
        :return component notes of chord
        """
        if self._on:
            self._quality.append_on_chord(self.on, self.root)

        return self._quality.get_components(root=self._root, visible=visible)

    def _parse(self, chord):
        """ parse a chord

        :param str chord: Name of chord.
        """
        root, quality, appended, on = parse(chord)
        self._root = root
        self._quality = quality
        self._appended = appended
        self._on = on

    def _reconfigure_chord(self):
        # TODO: Use appended
        self._chord = "{}{}{}{}".format(self._root,
                                        self._quality._quality,
                                        display_appended(self._appended),
                                        display_on(self._on))


def as_chord(chord):
    """ convert from str to Chord instance if input is str

    :type chord: str|pychord.Chord
    :param chord: Chord name or Chord instance
    :rtype: pychord.Chord
    :return: Chord instance
    """
    if isinstance(chord, Chord):
        return chord
    elif isinstance(chord, str):
        return Chord(chord)
    else:
        raise TypeError("input type should be str or Chord instance.")



# -*- coding: utf-8 -*-

import unittest

from pychord import Chord


class TestChordCreations(unittest.TestCase):

    def test_transpose_zero(self):
        c = Chord("Am")
        c.transpose(0)
        self.assertEqual(c.root, "A")
        self.assertEqual(c.quality.quality, "m")
        self.assertEqual(c, Chord("Am"))

    def test_transpose_positive(self):
        c = Chord("Am")
        c.transpose(3)
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "m")
        self.assertEqual(c, Chord("Cm"))

    def test_transpose_negative(self):
        c = Chord("Am")
        c.transpose(-4)
        self.assertEqual(c.root, "F")
        self.assertEqual(c.quality.quality, "m")
        self.assertEqual(c, Chord("Fm"))

    def test_transpose_slash(self):
        c = Chord("Am7/G")
        c.transpose(3)
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "m7")
        self.assertEqual(c.on, "Bb")
        self.assertEqual(c, Chord("Cm7/Bb"))

    def test_invalid_transpose_type(self):
        c = Chord("Am")
        self.assertRaises(TypeError, c.transpose, "A")

    def test_transpose_eq1(self):
        c = Chord("C")
        c.transpose(1)
        self.assertEqual(c, Chord("C#"))
        self.assertEqual(c, Chord("Db"))

    def test_transpose_eq2(self):
        c = Chord("C")
        c.transpose(2)
        self.assertEqual(c, Chord("D"))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .chord import Chord
from .constants.qualities import QUALITY_DICT
from .utils import note_to_val


def note_to_chord(notes):
    """ Convert note list to chord list

    :param list[str] notes: list of note arranged from lower note. ex) ["C", "Eb", "G"]
    :rtype: list[pychord.Chord]
    :return: list of chord
    """
    if not notes:
        raise ValueError("Please specify notes which consist a chord.")
    root = notes[0]
    root_and_positions = []
    for rotated_notes in get_all_rotated_notes(notes):
        rotated_root = rotated_notes[0]
        root_and_positions.append([rotated_root, notes_to_positions(rotated_notes, rotated_notes[0])])
    chords = []
    for temp_root, positions in root_and_positions:
        quality = find_quality(positions)
        if quality is None:
            continue
        if temp_root == root:
            chord = "{}{}".format(root, quality)
        else:
            chord = "{}{}/{}".format(temp_root, quality, root)
        chords.append(Chord(chord))
    return chords


def notes_to_positions(notes, root):
    """ Get notes positions.

    ex) notes_to_positions(["C", "E", "G"], "C") -> [0, 4, 7]

    :param list[str] notes: list of notes
    :param str root: the root note
    :return list[int]: list of note positions
    """
    root_pos = note_to_val(root)
    current_pos = root_pos
    positions = []
    for note in notes:
        note_pos = note_to_val(note)
        if note_pos < current_pos:
            note_pos += 12 * ((current_pos - note_pos) // 12 + 1)
        positions.append(note_pos - root_pos)
        current_pos = note_pos
    return positions


def get_all_rotated_notes(notes):
    """ Get all rotated notes

    get_all_rotated_notes([1,3,5]) -> [[1,3,5],[3,5,1],[5,1,3]]

    :type notes: list[str]
    :rtype: list[list[str]]
    """
    notes_list = []
    for x in range(len(notes)):
        notes_list.append(notes[x:] + notes[:x])
    return notes_list


def find_quality(positions):
    """ Find a quality consists of positions

    :param list[int] positions: note positions
    :rtype: str|None
    """
    for q, p in QUALITY_DICT.items():
        if positions == list(p):
            return q
    return None



# -*- coding: utf-8 -*-

import unittest

from pychord import Chord


class TestChordCreations(unittest.TestCase):

    def test_normal_chord(self):
        c = Chord("C")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [0, 4, 7])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["C", "E", "G"])

    def test_minor_chord(self):
        c = Chord("Am")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [9, 12, 16])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["A", "C", "E"])

    def test_dim_chord(self):
        c = Chord("Ddim")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [2, 5, 8])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["D", "F", "G#"])

    def test_aug_chord(self):
        c = Chord("Eaug")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [4, 8, 12])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["E", "G#", "C"])

    def test_slash_chord(self):
        c = Chord("CM9/D")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [-10, 0, 4, 7, 11])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["D", "C", "E", "G", "B"])

    def test_sus4_chord(self):
        c = Chord("Fsus4")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [5, 10, 12])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["F", "Bb", "C"])

    def test_seventh_chord(self):
        c = Chord("G7")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [7, 11, 14, 17])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["G", "B", "D", "F"])

    def test_sixth_chord(self):
        c = Chord("C6")
        com0 = c.components(visible=False)
        self.assertEqual(com0, [0, 4, 7, 9])
        com1 = c.components(visible=True)
        self.assertEqual(com1, ["C", "E", "G", "A"])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from .constants import QUALITY_DICT
from .utils import note_to_val, val_to_note


class Quality(object):
    """ Chord quality

    :param str _quality: str expression of chord quality
    """
    def __init__(self, quality):
        """ Constructor of chord quality

        :param str quality: name of quality
        """
        if quality not in QUALITY_DICT:
            raise ValueError("unknown quality {}".format(quality))
        self._quality = quality
        self.components = list(QUALITY_DICT[quality])

    def __unicode__(self):
        return self._quality

    def __str__(self):
        return self._quality

    def __eq__(self, other):
        if not isinstance(other, Quality):
            raise TypeError("Cannot compare Quality object with {} object".format(type(other)))
        return self._quality == other.quality

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def quality(self):
        """ Get name of quality """
        return self._quality

    def get_components(self, root='C', visible=False):
        """ Get components of chord quality

        :param str root: the root note of the chord
        :param bool visible: returns the name of notes if True
        :rtype: list[(str or int)]
        :return components of chord quality
        """
        root_val = note_to_val(root)
        components = [v + root_val for v in self.components]

        if visible:
            components = [val_to_note(c, scale=root) for c in components]

        return components

    def append_on_chord(self, on_chord, root):
        """ Append on chord

        To create Am7/G
        q = Quality('m7')
        q.append_on_chord('G', root='A')

        :param str on_chord: bass note of the chord
        :param str root: root note of the chord
        """
        root_val = note_to_val(root)
        on_chord_val = note_to_val(on_chord) - root_val

        list_ = list(self.components)
        for idx, val in enumerate(list_):
            if val % 12 == on_chord_val:
                self.components.remove(val)
                break

        if on_chord_val > root_val:
            on_chord_val -= 12

        if on_chord_val not in self.components:
            self.components.insert(0, on_chord_val)

    def append_note(self, note, root, scale=0):
        """ Append a note to quality

        :param str note: note to append on quality
        :param str root: root note of chord
        :param int scale: key scale
        """
        root_val = note_to_val(root)
        note_val = note_to_val(note) - root_val + scale * 12
        if note_val not in self.components:
            self.components.append(note_val)
            self.components.sort()

    def append_notes(self, notes, root, scale=0):
        """ Append notes to quality

        :param list[str] notes: notes to append on quality
        :param str root: root note of chord
        :param int scale: key scale
        """
        for note in notes:
            self.append_note(note, root, scale)



# -*- coding: utf-8 -*-

import unittest

from pychord import Chord
from pychord.chord import as_chord


class TestChordCreations(unittest.TestCase):

    def test_normal_chord(self):
        c = Chord("C")
        self.assertEqual(c.root, "C")
        self.assertEqual(c.quality.quality, "")

    def test_invalid_chord(self):
        self.assertRaises(ValueError, Chord, "H")

    def test_minor_chord(self):
        c = Chord("Am")
        self.assertEqual(c.root, "A")
        self.assertEqual(c.quality.quality, "m")

    def test_m75_chord(self):
        c = Chord("Bm7-5")
        self.assertEqual(c.root, "B")
        self.assertEqual(c.quality.quality, "m7-5")

    def test_invalid_quality_chord(self):
        self.assertRaises(ValueError, Chord, "Csus3")

    def test_slash_chord(self):
        c = Chord("F/G")
        self.assertEqual(c.root, "F")
        self.assertEqual(c.quality.quality, "")
        self.assertEqual(c.on, "G")

    def test_minor_slash_chord(self):
        c = Chord("Dm/G")
        self.assertEqual(c.root, "D")
        self.assertEqual(c.quality.quality, "m")
        self.assertEqual(c.on, "G")

    def test_invalid_slash_chord(self):
        self.assertRaises(ValueError, Chord, "C/H")

    def test_eq(self):
        c1 = Chord("C")
        c2 = Chord("C")
        self.assertEqual(c1, c2)

    def test_invalid_eq(self):
        c = Chord("C")
        with self.assertRaises(TypeError):
            print(c == 0)


class TestAsChord(unittest.TestCase):

    def test_as_chord_chord_input(self):
        c = Chord("C")
        ac = as_chord(c)
        self.assertEqual(c, ac)

    def test_as_chord_str_input(self):
        chord = "C"
        c = Chord(chord)
        ac = as_chord(chord)
        self.assertEqual(c, ac)

    def test_as_chord_invalid_input(self):
        with self.assertRaises(TypeError):
            as_chord(1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .quality import Quality
from .constants import QUALITY_DICT
from .utils import NOTE_VAL_DICT


def parse(chord):
    """ Parse a string to get chord component

    :param str chord: str expression of a chord
    :rtype: (str, pychord.Quality, str, str)
    :return (root, quality, appended, on):
    """
    if len(chord) > 1 and chord[1] in ("b", "#"):
        root = chord[:2]
        rest = chord[2:]
    else:
        root = chord[:1]
        rest = chord[1:]
    check_note(root, chord)
    on_chord_idx = rest.find("/")
    if on_chord_idx >= 0:
        on = rest[on_chord_idx + 1:]
        rest = rest[:on_chord_idx]
        check_note(on, chord)
    else:
        on = None
    if rest in QUALITY_DICT:
        quality = Quality(rest)
    else:
        raise ValueError("Invalid chord {}: Unknown quality {}".format(chord, rest))
    # TODO: Implement parser for appended notes
    appended = []
    return root, quality, appended, on


def check_note(note, chord):
    """ Return True if the note is valid.

    :param str note: note to check its validity
    :param str chord: the chord which includes the note
    :return bool:
    """
    if note not in NOTE_VAL_DICT:
        raise ValueError("Invalid chord {}: Unknown note {}".format(chord, note))
    return True



# -*- coding: utf-8 -*-

import unittest

from pychord import Quality


class TestQuality(unittest.TestCase):

    def test_eq(self):
        c1 = Quality("m7-5")
        c2 = Quality("m7-5")
        self.assertEqual(c1, c2)

    def test_invalid_eq(self):
        c = Quality("m7")
        with self.assertRaises(TypeError):
            print(c == 0)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from .constants import NOTE_VAL_DICT, SCALE_VAL_DICT


def note_to_val(note):
    """ Convert note to int

    >>> note_to_val("C")
    0
    >>> note_to_val("B")
    11

    :type note: str
    :rtype: int
    """
    if note not in NOTE_VAL_DICT:
        raise ValueError("Unknown note {}".format(note))
    return NOTE_VAL_DICT[note]


def val_to_note(val, scale="C"):
    """ Convert int to note

    >>> val_to_note(0)
    "C"
    >>> val_to_note(11, "D")
    "D#"

    :type val: int
    :param str scale: key scale
    :rtype: str
    """
    val %= 12
    return SCALE_VAL_DICT[scale][val]


def transpose_note(note, transpose, scale="C"):
    """ Transpose a note

    :param str note: note to transpose
    :type transpose: int
    :param str scale: key scale
    :rtype: str
    :return: transposed note
    """
    val = note_to_val(note)
    val += transpose
    return val_to_note(val, scale)


def display_appended(appended):
    # TODO: Implement this
    return ""


def display_on(on_note):
    if on_note:
        return "/{}".format(on_note)
    return ""



# -*- coding: utf-8 -*-

import unittest

from pychord import Chord
from pychord.analyzer import get_all_rotated_notes, note_to_chord, notes_to_positions


class TestNotesToPositions(unittest.TestCase):

    def test_one(self):
        pos = notes_to_positions(["C"], "C")
        self.assertEqual(pos, [0])

    def test_power(self):
        pos = notes_to_positions(["C", "G"], "C")
        self.assertEqual(pos, [0, 7])

    def test_major(self):
        pos = notes_to_positions(["D", "F#", "A"], "D")
        self.assertEqual(pos, [0, 4, 7])

    def test_seventh(self):
        pos = notes_to_positions(["E", "G#", "B", "D"], "E")
        self.assertEqual(pos, [0, 4, 7, 10])

    def test_add9(self):
        pos = notes_to_positions(["Ab", "C", "Eb", "Bb"], "Ab")
        self.assertEqual(pos, [0, 4, 7, 14])

    def test_ninth(self):
        pos = notes_to_positions(["F", "A", "C", "Eb", "G"], "F")
        self.assertEqual(pos, [0, 4, 7, 10, 14])

    def test_eleventh(self):
        pos = notes_to_positions(["G", "B", "D", "F", "A", "C"], "G")
        self.assertEqual(pos, [0, 4, 7, 10, 14, 17])

    def test_thirteenth(self):
        pos = notes_to_positions(["A", "C#", "E", "G", "B", "D", "F#"], "A")
        self.assertEqual(pos, [0, 4, 7, 10, 14, 17, 21])


class TestGetAllRotatedNotes(unittest.TestCase):

    def test_two(self):
        notes_list = get_all_rotated_notes(["C", "G"])
        self.assertEqual(notes_list, [["C", "G"], ["G", "C"]])

    def test_three(self):
        notes_list = get_all_rotated_notes(["C", "F", "G"])
        self.assertEqual(notes_list, [["C", "F", "G"], ["F", "G", "C"], ["G", "C", "F"]])


class TestNoteToChord(unittest.TestCase):

    def test_major(self):
        chords = note_to_chord(["C", "E", "G"])
        self.assertEqual(chords, [Chord("C")])

    def test_major_on_third(self):
        chords = note_to_chord(["F#", "A", "D"])
        self.assertEqual(chords, [Chord("D/F#")])

    def test_major_on_fifth(self):
        chords = note_to_chord(["B", "E", "G#"])
        self.assertEqual(chords, [Chord("E/B")])

    def test_dim(self):
        chords = note_to_chord(["Eb", "Gb", "A"])
        self.assertEqual(chords, [Chord("Ebdim")])

    def test_dim6(self):
        chords = note_to_chord(["Eb", "Gb", "A", "C"])
        self.assertEqual(chords, [Chord("Ebdim6"), Chord("Gbdim6/Eb"), Chord("Adim6/Eb"), Chord("Cdim6/Eb")])

    def test_aug(self):
        chords = note_to_chord(["F", "A", "Db"])
        self.assertEqual(chords, [Chord("Faug"), Chord("Aaug/F"), Chord("Dbaug/F")])

    def test_add9(self):
        chords = note_to_chord(["C", "E", "G", "D"])
        self.assertEqual(chords, [Chord("Cadd9")])

    def test_m7dim5(self):
        chords = note_to_chord(["F#", "A", "C", "E"])
        self.assertEqual(chords, [Chord("F#m7-5")])

    def test_call_repeatedly(self):
        for _ in range(2):
            chords = note_to_chord(["Eb", "Gb", "A", "C"])
            self.assertEqual(chords, [Chord("Ebdim6"), Chord("Gbdim6/Eb"), Chord("Adim6/Eb"), Chord("Cdim6/Eb")])
            self.assertEqual(chords[0].components(visible=True), ["Eb", "Gb", "A", "C"])



from setuptools import setup, find_packages
version = '0.3.2'

setup(
    name='pychord',
    version=version,
    description="A library to handle musical chords in python.",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        ],
    keywords='music chord',
    author='Yuma Mihira',
    author_email='info@yurax2.com',
    url='https://github.com/yuma-m/pychord',
    license='MIT',
    packages=find_packages(exclude=['test']),
    include_package_data=True,
    zip_safe=True,
    long_description=open('README.rst').read(),
    test_suite='test',
)
